# Rhetos.ElasticSearch

Rhetos.ElasticSearch is a DSL package (a plugin module) for [Rhetos development platform](https://github.com/Rhetos/Rhetos).
It provides an implementation of filtering data via external ElasticSearch engine.

Contents:

1. [Installation and configuration](#installation-and-configuration)
2. [Usage](#usage)
   1. [Creating ElasticSearch index](#creating-elasticsearch-index)
   2. [Creating search function](#creating-search-function)
   3. [Command Line Interface](#command-line-interface)
   4. [RhetosElasticClient](#rhetoselasticclient)

See [rhetos.org](http://www.rhetos.org/) for more information on Rhetos.

## Installation and configuration

To install this package to a Rhetos server, add it to the Rhetos server's *RhetosPackages.config* file
and make sure the NuGet package location is listed in the *RhetosPackageSources.config* file.

* The package ID is "**Rhetos.ElasticSearch**".
  This package is available at the [NuGet.org](https://www.nuget.org/) online gallery.
  The Rhetos server can install the package directly from there, if the gallery is listed in *RhetosPackageSources.config* file.
* For more information, see [Installing plugin packages](https://github.com/Rhetos/Rhetos/wiki/Installing-plugin-packages).

Configuration of the plugin is done in `rhetos-app.settings.json`, like this:

```json
{
  "Rhetos": {
    "ElasticSearch": {
      "ServerUrls": ["https://ElasticSearchServer1", "https://ElasticSearchServer2"], //Mandatory parameter. List of Urls of the ElasticSearch server instances.
      "UserName": "admin", //Optional parameter. If provided in combination with password, BasicAuthentication will be used.
      "Password": "admin", //Optional parameter. 
      "EnvironmentOverride": "ElasticSearchTest", //Optional parameter. Name of the application instance. Used for constructing index name. If not specified environmet is constructed from server and database name retireved from Rhetos connection string.
      "DefaultQuerySize": 20, //This is default value of the parameter.
      "MaxQuerySize": 1000, //This is default value of the parameter.
      "IndexSaveWaitSeconds": 300, //This is default value of the parameter.
      "IndexElasticBatchSize": 20000, //This is default value of the parameter.
      "NumberOfShards": 0, //This is default value of the parameter. Value 0 means that number of shards will be taken from number of nodes, otherwise number of shards will be as value provided.
      "NumberOfReplicas": 0, //This is default value of the parameter. For production purposes you will want to have at least 1 replica.
      "PatternAnalyzerDataRegex": "\d\w*([/\-,\.]\d\w*)+|\w+", //This is default value of the parameter. Regex used for pattern analyzer in elastic index.
      "PatternAnalyzerFilterRegex": "\d\w*([/\-,\.]\d\w*)+|\w+", //This is default value of the parameter. Regex used for preprocesing filter on property that has pattern analyzer defined.
    }
  }
}
```

[ICU Analysis Plugin](https://www.elastic.co/guide/en/elasticsearch/plugins/current/analysis-icu.html) needs to be installed on the Elasticsearch server.

## Usage

For the example purposes following entity will be used:

```cs
Module ElasticTest
{
  Entity Catalog
  {
    ShortString Name;
  }
  
  Entity TheEntity
  {
    ShortString Name;
    LongString Description;
    Reference Catalog;
    Date TheDate;
    LongString SubItemIds;
    Integer TheInt;
    Decimal TheDecimal;
    Money TheMoney;
    Bool TheBool;
    DateTime TheDateTime;
  }
}
```

### Creating ElasticSearch index

**ElasticIndex** concept is used to create data source for populating index. It also creates:

* `Entity` **[ElasticIndexName]Queue** - represents a queue for syncing data to index in ElasticSearch.
* `Action` **[ElasticIndexName]QueueAllItems** - queues all items into queue in order to reindex all of the data.
* `Action` **[ElasticIndexName]SyncIndex** - creates index in ElasticSearch (if not allready created) and syncs queued data to the index.
* `Action` **[ElasticIndexName]SyncItems** - creates index in ElasticSearch (if not allready created) and syncs items which ids are passed in the `ItemIds` parameter of the action. Action also removes passed `ItemIds` from **[ElasticIndexName]Queue** on success. 

**ElasticIndex** concept is derived from **Browse** concept so you can use **Take** concepts for adding wanted properties. For each property you can define how that property will be searched. For that purpose number of 'ElasticWhere' concepts are defined:

* **ElasticWhere** - this concept is used for **ShortString**, **LongString** and **Bool** properties.
* **ElasticWhereAny** - this concept is used for **Guid** and **Reference** properties.
* **ElasticWhereAll** - this concept is used for **LongString** properties that contains number of guids separated by spaces.
* **ElasticWhereRange** - this concept is used for date and number properties.

Here is the example of creating ElasticSearch index:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
	QueueOrder "TheDateTime DESC"; // Optional sort for queue when indexing entire elastic index.
	
    Take Name { ElasticWhere; }
    Take Description; //Just add property to index. We will add custom search for this property later.
    Take Catalog { ElasticWhereAny; }
    Take CatalogName 'Catalog.Name' { ElasticWhere; } 
    Take TheDate { ElasticWhereRange; }
    Take TheDateTime { ElasticWhereRange; }
    Take SubItemIds { ElasticWhereAll;  }
    Take TheDecimal { ElasticWhereRange; }
    Take TheMoney { ElasticWhereRange; }
    Take TheInt { ElasticWhereRange; }
    Take TheBool { ElasticWhere; }
  }
```

**ElasticIndex** concept also has built in functionality of automatically inserting items into queue when source item is inserted or updated. If you want additional dependency entities to be included in this functionality simply add **ChangesOnChangedItems** accordingly:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
    Take Name { ElasticWhere; }
    Take Description; //Just add property to index. We will add custom search for this property later.
    Take Catalog { ElasticWhereAny; }
    Take CatalogName 'Catalog.Name' { ElasticWhere; } 
    Take TheDate { ElasticWhereRange; }
    Take TheDateTime { ElasticWhereRange; }
    Take SubItemIds { ElasticWhereAll;  }
    Take TheDecimal { ElasticWhereRange; }
    Take TheMoney { ElasticWhereRange; }
    Take TheInt { ElasticWhereRange; }
    Take TheBool { ElasticWhere; }

    ChangesOnChangedItems ElasticTest.Catalog 'FilterCriteria' 'changedItems =>
      {
        var ids = changedItems.Select(item => item.ID).Distinct();
        return new FilterCriteria("CatalogID", "in", ids);
      }';
  }
```

For **ShortString** and **LongString** properties you can use number of analyzers for populating elastic index. You can use any of ElasticSearch built-in analyzers. There are also two additional analyzers provided:

* **whitespace_lowercase** - uses built-in whitespace tokenizer and converts text to lowercase
* **pattern_lowercase** - uses built-in pattern tokenizer (with regex provided in configuration file - `ElasticSearch:PatternAnalyzerDataRegex`) and converts text to lowercase. This analyzer also implements IFilterParser interface which is called for filter string preprocessing with regex provided in configuration file (`ElasticSearch:PatternAnalyzerFilterRegex`)

You can asign analyzer for certain property like this:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
    Take Description { ElasticWhere; ElasticAnalyzer "standard"; } //you can use any of the Elastic built in or custom analyzers
  }
```

You can also register your own analyzer and then use it. Here is an example of custom analyzers:

```cs
  public class WhitespaceLowerCaseAnalyzer : ICustomElasticAnalyzer
  {
    public string Name => "whitespace_lowercase";
    public Func<CustomAnalyzerDescriptor, ICustomAnalyzer> Analyzer => descriptor => descriptor.Tokenizer("whitespace").Filters("lowercase");
    public Func<TokenizersDescriptor, TokenizersDescriptor> Tokenizer => null;
  }
```

or like this when your analyzer also needs to preprocess filter string

```cs
  public class CustomPatternAnalyzer : ICustomElasticAnalyzer, IFilterParser
  {
    public string Name => "custom_pattern_analyzer";
    public Func<CustomAnalyzerDescriptor, ICustomAnalyzer> Analyzer => descriptor => descriptor.Tokenizer("custom_pattern_tokenizer").Filters("lowercase");
    public Func<TokenizersDescriptor, TokenizersDescriptor> Tokenizer => descriptor =>  
      descriptor.Pattern("custom_pattern_tokenizer", p => p.Pattern(@"\d\w*([/\-,\.]\d\w*)+|UP/II?[/\-\d]*|\w+").Group(0));
	  
	public string ParseFilter(string filter)
    {
      if (string.IsNullOrWhiteSpace(filter))
        return filter;
      var tokensRegex = new Regex(@"\d\w*([/\-,\.]\d\w*)+|UP/II?[/\-\d]*|\w+", RegexOptions.IgnoreCase);
      return string.Join(" ", tokensRegex.Matches(filter).Cast<Match>().Select(x => x.Value).ToList());
    }
  }
```

Once you have created your analyzer you have to register it in Autofac:

```cs
  [Export(typeof(Module))]
  public class AutofacModuleConfiguration : Module
  {
    protected override void Load(ContainerBuilder builder)
    {
      builder.RegisterType<WhitespaceLowerCaseAnalyzer>().As<ICustomElasticAnalyzer>().SingleInstance();
      builder.RegisterType<CustomPatternAnalyzer>().As<ICustomElasticAnalyzer>().SingleInstance();
      //When analyzer implements IFilterParser interface it also needs to be registered directly  
      builder.RegisterType<CustomPatternAnalyzer>().SingleInstance();
      base.Load(builder);
    }
  }
```

And now you can us it in your DSL:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
    Take Description { ElasticWhere; ElasticAnalyzer "custom_pattern_analyzer"; } //you can use any of the Elastic built in analyzers
  }
```

### Creating search function

Once **ElasticIndex** is created it can be used on any **Browse** or **SqlQueryable** for the same source **Entity**. To create search function **ElasticSearch** concept is used:

```cs
  Browse TheEntityBrowse ElasticTest.TheEntity
  {
    Take Name;
    Take Description;
    Take Catalog;
    Take CatalogName 'Catalog.Name';
    Take TheDate;
    
    ElasticSearch ElasticTest.TheEntityElastic;
  }
```

**ElasticSearch** concept creates search function named `[BrowseName]Search`. For each property in ElasticIndex that has 'ElasticWhere' defined, specific function parameter(s) are created:

* **ElasticWhere** - parameter with same name and type is created
* **ElasticWhereAny** - parameter with name [PropertyName]List and type List<Guid?> is created.
* **ElasticWhereAll** - parameter with name [PropertyName]IDList and type List<Guid?> is created.
* **ElasticWhereRange** - two parameters of the same type and names [PropertyName]From and [PropertyName]To are created.

In each search function there is also **Ids** property of type List<Guid?> in order to retrieve only exact items.

Along with search parameters, sorting and paging parameters are created:

* **Sort** - string, suports property name and sort direction, for example `"Name"` or `"TheDate DESC"`
* **Skip** - integer
* **Take** - integer

**ElasticSearch** concept also creates return type for search function named `[BrowseName]SearchResult`. Return type is a **DataStructure** with following properties:

* **Data** - `List<TheEntityBrowse>` for the example above.
* **Total** - `int` that represents number of items which satisfies search parameters.

Now we will add custom search for Description property for example above. For that purpose **AdditionalElasticQuery** concept is used.

```cs

  ShortString TheEntityBrowseSearch TheDescription; //add parameter to search function
  
  Browse TheEntityBrowse ElasticTest.TheEntity
  {
    ElasticSearch ElasticTest.TheEntityElastic
    {
      AdditionalElasticQuery 'Custom additional filter' '
        if (!string.IsNullOrWhiteSpace(parameter.TheDescription))
          query.Where(q => q.Where(f => f.Description, parameter.TheDescription));
      ';      
    }
  }
```

In your custom code you can use following provided variables:

* **parameter** - instance of search function parameters
* **repository** - instance of DomRepository
* **context** - instance of ExecutionContext
* **userInfo** - instance of IUserInfo
* **query** - instance of elastic query builder

One of the main purposes of custom search functionality is implementation of row permissions, since using **RowPermissions** concepts with **ElasticSearch** concept is impossible. Idea is to put data used for row permissions directly into index and then filter by that data. Here is an example:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
    Take Name { ElasticWhere; }
    Take CatalogID;
    Take CatalogName { ElasticWhere; }
    Take TheDate { ElasticWhereRange; }
    Take AllowedUsers 'Base.Extension_RowPermissionsQueryable.AllowedUsers' { ElasticWhereAll; }
  }

  Browse TheEntityBrowse ElasticTest.TheEntity
  {
    Take Name;
    Take Description;
    Take CatalogName 'Catalog.Name';
    Take TheDate;

    ElasticSearch ElasticTest.TheEntityElastic
    {
      AdditionalElasticQuery 'Row Permissions' '
        var userId = repository.Common.Principal.Query(x => x.Name == context.UserInfo.UserName).Single().ID;
        var allowedCatalogIds = repository.ElasticTest.AllowedCatalogs.Query(x => x.UserID == userId).Select(x => x.CatalogID).ToList();
        query.Where(q => q.WhereAll(f => f.AllowedUsers, new[]{userId}) || q.WhereAny(f => f.CatalogID), allowedCatalogIds);
      ';      
    }
  }
```

**ElasticSearch** concept can also be used on **ElasticIndex** directly. That way additional sql query to retrieve filtered data is ommited since it is the same data that is stored in ElasticSearch. Drawback of this approach is that when new property is added to the **ElasticIndex** entire index must be reindexed in order to return that new property. Here is an example:

```cs
  ElasticIndex TheEntityElastic ElasticTest.TheEntity
  {
    Take Name { ElasticWhere; }
    Take Description; //We do not want to search this property, just to return it in the result.
    Take Catalog { ElasticWhereAny; }
    Take CatalogName 'Catalog.Name' { ElasticWhere; } 
    Take TheDate { ElasticWhereRange; }

    ElasticSearch ElasticTest.TheEntityElastic;
  }
```

### Command Line Interface

For manual indexing/reindexing CLI tool is provided. **esc.exe** is placed in `Plugin` folder of the Rhetos server. CLI switches are:

* **/Index:[NameOfElasticIndex]** or **/I:[NameOfElasticIndex]** - name of the index that should be used. For example `/I:MyModule.MyElasticIndex`
* **/Del** or **/D** - deletes index in ElasticSearch.
* **/Queue** or **/Q** - queues all the items from the source to queue for syncing to ElasticSearch.
* **/Sync** or **/S** - syncs all queued items to ElasticSearch.
* **/Reindex** or **/R** - queues and syncs all items to ElasticSearch. This command is convenient for consecutive call of above three commands.
* **/V** - prints additional data in console during long operations.

### RhetosElasticClient

In **RhetosElasticClient** class is key implementation for communication with ElasticSearch engine. Class is registered in Rhetos and you can use it in your code for example in your DOM code:

```cs
	var elasticClient = container.Resolve<Rhetos.ElasticSearch.RhetosElasticClient>();
```

Or you can use it in your DSL like this:

```cs
Module Common
{
  Action ElasticIndexItems
  '(parameters, repository, userInfo) =>
  {
    _elasticClient.IndexItems(parameters.IndexName, parameters.ItemIds);
  }'
  {
    ShortString IndexName;
    ListOf Guid ItemIds;
    RepositoryUses '_elasticClient' 'Rhetos.ElasticSearch.RhetosElasticClient, Rhetos.ElasticSearch';
  }
}
```

Methods provided in **RhetosElasticClient** are:

* `SearchResult<T> Search<T>(ElasticQuery<T> query)` - method for searching the ElasticSearch index.
* `IndexItems(string elasticIndexName, IEnumerable<Guid> ids)` - method for indexing items with specified ids. There are also following extension methods provided: `IndexItems(string elasticIndexName, IEnumerable<Guid?> ids)`,  `IndexItems<TIndex>(IEnumerable<Guid> ids)` and `IndexItems<TIndex>(IEnumerable<Guid?> ids)`
*  `IndexQueuedItems(string elasticIndexName, int? batchSize = null)` - method for indexing one batch of items that are queued. If batchSize is omitted the value of configuration parameter **ElasticSearch:IndexRhetosBatchSize** is used. There is also extension method provided `IndexQueuedItems<TIndex>(int? batchSize = null)`
* `DeleteIndex(string elasticIndexName)` - method deletes entire index from ElasticSearch. Use this method only for testing purposes or when structure of index is dramaticaly changed. There is also extension method provided `DeleteIndex<TIndex>()`